@@ -1,5 +1,4 @@
-use core::num::{NonZeroU32, NonZeroU64};
-
+use core::num::NonZeroU64;
 use crate::backend::proxy::ContextFunction;
 use crate::backend::util::GpuImage;
 use crate::backend::util::{ToStd140, UniformsBuffer};
@@ -297,7 +296,7 @@ impl Context {
 
 		let surface = unsafe { self.instance.create_surface(&window) };
 		let size = glam::UVec2::new(window.inner_size().width, window.inner_size().height);
-		let swap_chain = create_swap_chain(size, &surface, self.swap_chain_format, &self.device);
+		configure_surface(size, &surface, self.swap_chain_format, &self.device);
 		let uniforms = UniformsBuffer::from_value(&self.device, &WindowUniforms::no_image(), &self.window_bind_group_layout);
 
 		let window = Window {
@@ -305,7 +304,6 @@ impl Context {
 			preserve_aspect_ratio: options.preserve_aspect_ratio,
 			background_color: options.background_color,
 			surface,
-			swap_chain,
 			uniforms,
 			image: None,
 			user_transform: Affine2::IDENTITY,
@@ -346,7 +344,7 @@ impl Context {
 			.find(|w| w.id() == window_id)
 			.ok_or(InvalidWindowId { window_id })?;
 
-		window.swap_chain = create_swap_chain(new_size, &window.surface, self.swap_chain_format, &self.device);
+		configure_surface(new_size, &window.surface, self.swap_chain_format, &self.device);
 		window.uniforms.mark_dirty(true);
 		Ok(())
 	}
@@ -365,9 +363,9 @@ impl Context {
 		};
 
 		let frame = window
-			.swap_chain
-			.get_current_frame()
-			.expect("Failed to acquire next swap chain texture");
+			.surface
+			.get_current_texture()
+			.expect("Failed to acquire next frame");
 
 		let mut encoder = self.device.create_command_encoder(&Default::default());
 
@@ -383,7 +381,7 @@ impl Context {
 			&window.uniforms,
 			image,
 			Some(window.background_color),
-			&frame.output.view,
+			&frame.texture.create_view(&wgpu::TextureViewDescriptor::default()),
 		);
 		if window.overlays_visible {
 			for overlay in &window.overlays {
@@ -393,15 +391,19 @@ impl Context {
 					&window.uniforms,
 					overlay,
 					None,
-					&frame.output.view,
+					&frame.texture.create_view(&wgpu::TextureViewDescriptor::default()),
 				);
 			}
 		}
 		self.queue.submit(std::iter::once(encoder.finish()));
+		frame.present();
 		Ok(())
 	}
 
+	#[cfg(feature = "save")]
 	fn render_to_texture(&self, window_id: WindowId, overlays: bool) -> Result<Option<(String, crate::BoxImage)>, InvalidWindowId> {
+		use core::num::NonZeroU32;
+
 		let window = self
 			.windows
 			.iter()
@@ -414,6 +416,7 @@ impl Context {
 		};
 
 		let bytes_per_row = align_next_u32(image.info().size.x * 4, wgpu::COPY_BYTES_PER_ROW_ALIGNMENT);
+		let width_scale = image.info().size.x as f32 * 4.0 / bytes_per_row as f32;
 
 		let size = wgpu::Extent3d {
 			width: div_round_up(bytes_per_row, 4),
@@ -422,14 +425,14 @@ impl Context {
 		};
 
 		let window_uniforms = WindowUniforms {
-			transform: Affine2::IDENTITY,
-			image_size: image.info().size.as_f32(),
+			transform: Affine2::from_scale([width_scale, 1.0].into()),
+			image_size: image.info().size.as_vec2(),
 		};
 		let window_uniforms = UniformsBuffer::from_value(&self.device, &window_uniforms, &self.window_bind_group_layout);
 
 		let target = self.device.create_texture(&wgpu::TextureDescriptor {
 			label: Some(&format!("{}_render", image.name())),
-			usage: wgpu::TextureUsage::RENDER_ATTACHMENT | wgpu::TextureUsage::COPY_SRC,
+			usage: wgpu::TextureUsages::RENDER_ATTACHMENT | wgpu::TextureUsages::COPY_SRC,
 			sample_count: 1,
 			mip_level_count: 1,
 			format: wgpu::TextureFormat::Rgba8Unorm,
@@ -467,7 +470,7 @@ impl Context {
 		let buffer = self.device.create_buffer(&wgpu::BufferDescriptor {
 			label: None,
 			size: u64::from(bytes_per_row) * u64::from(image.info().size.y),
-			usage: wgpu::BufferUsage::COPY_DST | wgpu::BufferUsage::MAP_READ,
+			usage: wgpu::BufferUsages::COPY_DST | wgpu::BufferUsages::MAP_READ,
 			mapped_at_creation: false,
 		});
 
@@ -476,6 +479,7 @@ impl Context {
 				texture: &target,
 				mip_level: 0,
 				origin: wgpu::Origin3d::ZERO,
+				aspect: wgpu::TextureAspect::All,
 			},
 			wgpu::ImageCopyBuffer {
 				buffer: &buffer,
@@ -584,7 +588,7 @@ impl Context {
 		let mut event_handlers = std::mem::take(&mut self.event_handlers);
 
 		let mut stop_propagation = false;
-		event_handlers.retain_mut(|handler| {
+		RetainMut::retain_mut(&mut event_handlers, |handler| {
 			if stop_propagation {
 				false
 			} else {
@@ -613,7 +617,7 @@ impl Context {
 		let mut event_handlers = std::mem::take(&mut self.windows[window_index].event_handlers);
 
 		let mut stop_propagation = false;
-		event_handlers.retain_mut(|handler| {
+		RetainMut::retain_mut(&mut event_handlers, |handler| {
 			if stop_propagation {
 				false
 			} else {
@@ -698,7 +702,7 @@ impl Context {
 	}
 }
 
-fn select_backend() -> wgpu::BackendBit {
+fn select_backend() -> wgpu::Backends {
 	let backend = std::env::var_os("WGPU_BACKEND").unwrap_or_else(|| "primary".into());
 	let backend = match backend.to_str() {
 		Some(backend) => backend,
@@ -709,19 +713,19 @@ fn select_backend() -> wgpu::BackendBit {
 	};
 
 	if backend.eq_ignore_ascii_case("primary") {
-		wgpu::BackendBit::PRIMARY
+		wgpu::Backends::PRIMARY
 	} else if backend.eq_ignore_ascii_case("vulkan") {
-		wgpu::BackendBit::VULKAN
+		wgpu::Backends::VULKAN
 	} else if backend.eq_ignore_ascii_case("metal") {
-		wgpu::BackendBit::METAL
+		wgpu::Backends::METAL
 	} else if backend.eq_ignore_ascii_case("dx12") {
-		wgpu::BackendBit::DX12
+		wgpu::Backends::DX12
 	} else if backend.eq_ignore_ascii_case("dx11") {
-		wgpu::BackendBit::DX11
+		wgpu::Backends::DX11
 	} else if backend.eq_ignore_ascii_case("gl") {
-		wgpu::BackendBit::GL
+		wgpu::Backends::GL
 	} else if backend.eq_ignore_ascii_case("webgpu") {
-		wgpu::BackendBit::BROWSER_WEBGPU
+		wgpu::Backends::BROWSER_WEBGPU
 	} else {
 		eprintln!("Unknown WGPU_BACKEND: {:?}", backend);
 		std::process::exit(1);
@@ -754,6 +758,7 @@ async fn get_device(instance: &wgpu::Instance) -> Result<(wgpu::Device, wgpu::Qu
 	let adapter = instance.request_adapter(&wgpu::RequestAdapterOptions {
 		power_preference: select_power_preference(),
 		compatible_surface: None, // TODO: can we use a hidden window or something?
+		force_fallback_adapter: false,
 	});
 
 	let adapter = adapter.await.ok_or(NoSuitableAdapterFound)?;
@@ -779,7 +784,7 @@ fn create_window_bind_group_layout(device: &wgpu::Device) -> wgpu::BindGroupLayo
 		label: Some("window_bind_group_layout"),
 		entries: &[wgpu::BindGroupLayoutEntry {
 			binding: 0,
-			visibility: wgpu::ShaderStage::VERTEX,
+			visibility: wgpu::ShaderStages::VERTEX,
 			count: None,
 			ty: wgpu::BindingType::Buffer {
 				ty: wgpu::BufferBindingType::Uniform,
@@ -797,7 +802,7 @@ fn create_image_bind_group_layout(device: &wgpu::Device) -> wgpu::BindGroupLayou
 		entries: &[
 			wgpu::BindGroupLayoutEntry {
 				binding: 0,
-				visibility: wgpu::ShaderStage::FRAGMENT,
+				visibility: wgpu::ShaderStages::FRAGMENT,
 				count: None,
 				ty: wgpu::BindingType::Buffer {
 					ty: wgpu::BufferBindingType::Uniform,
@@ -807,7 +812,7 @@ fn create_image_bind_group_layout(device: &wgpu::Device) -> wgpu::BindGroupLayou
 			},
 			wgpu::BindGroupLayoutEntry {
 				binding: 1,
-				visibility: wgpu::ShaderStage::FRAGMENT,
+				visibility: wgpu::ShaderStages::FRAGMENT,
 				count: None,
 				ty: wgpu::BindingType::Buffer {
 					ty: wgpu::BufferBindingType::Storage {
@@ -854,7 +859,7 @@ fn create_render_pipeline(
 						operation: wgpu::BlendOperation::Add,
 					},
 				}),
-				write_mask: wgpu::ColorWrite::ALL,
+				write_mask: wgpu::ColorWrites::ALL,
 			}],
 		}),
 		primitive: wgpu::PrimitiveState {
@@ -862,7 +867,7 @@ fn create_render_pipeline(
 			strip_index_format: None,
 			front_face: wgpu::FrontFace::Cw,
 			cull_mode: Some(wgpu::Face::Back),
-			clamp_depth: false,
+			unclipped_depth: false,
 			polygon_mode: wgpu::PolygonMode::Fill,
 			conservative: false,
 		},
@@ -872,25 +877,25 @@ fn create_render_pipeline(
 			mask: !0,
 			alpha_to_coverage_enabled: false,
 		},
+		multiview: None,
 	})
 }
 
 /// Create a swap chain for a surface.
-fn create_swap_chain(
+fn configure_surface(
 	size: glam::UVec2,
 	surface: &wgpu::Surface,
 	format: wgpu::TextureFormat,
 	device: &wgpu::Device,
-) -> wgpu::SwapChain {
-	let swap_chain_desc = wgpu::SwapChainDescriptor {
-		usage: wgpu::TextureUsage::RENDER_ATTACHMENT,
+) {
+	let config = wgpu::SurfaceConfiguration {
+		usage: wgpu::TextureUsages::RENDER_ATTACHMENT,
 		format,
 		width: size.x,
 		height: size.y,
 		present_mode: wgpu::PresentMode::Mailbox,
 	};
-
-	device.create_swap_chain(surface, &swap_chain_desc)
+	surface.configure(device, &config);
 }
 
 /// Perform a render pass of an image.
@@ -924,6 +929,7 @@ fn render_pass(
 	drop(render_pass);
 }
 
+#[cfg(feature = "save")]
 fn align_next_u32(input: u32, alignment: u32) -> u32 {
 	let remainder = input % alignment;
 	if remainder == 0 {
@@ -933,6 +939,7 @@ fn align_next_u32(input: u32, alignment: u32) -> u32 {
 	}
 }
 
+#[cfg(feature = "save")]
 fn div_round_up(input: u32, divisor: u32) -> u32 {
 	if input % divisor == 0 {
 		input / divisor
